import outlines.models as models
import outlines.text.generate as generate
import torch
from lark import Lark, Transformer, v_args
from lark.indenter import PythonIndenter
from transformers import (AutoModelForCausalLM, LogitsProcessorList, LogitsProcessor, AutoTokenizer)
import re
from abc import ABC,abstractmethod

model = models.transformers("gpt2")


class Completion(ABC):
    """
    A class used to generate completions when optimize.complete is called
    """

    @abstractmethod
    def complete(self, prompt):
        pass

    @staticmethod
    def create(regex, type, choices, grammar_type, *, default=None):
        completion = default
        if regex is not None:
            completion = GenerativeCompletion.regex_completion(regex)
        elif type is not None:
            completion = GenerativeCompletion.type_completion(type)
        elif choices is not None:
            completion = GenerativeCompletion.choices_completion(choices)
        elif grammar_type is not None:
            tokenizer = AutoTokenizer.from_pretrained("gpt2-medium", padding_side='left')
            completion = GrammarCompletion("gpt2-medium", tokenizer)
        return completion


class GenerativeCompletion(Completion):
    def __init__(self, generator, *generator_args):
        """
        Parameters: 
        -----------
        
        generator (Callable[[Transformers, ...generator_args], None]): Generator function to be used on the complete
        *generator_args (Any): Generator arguments (without model)
        
        """
        self.generator = generator
        self.generator_args = generator_args

    def complete(self, prompt):
        return self.generator(model, *self.generator_args)(prompt)

    @staticmethod
    def regex_completion(regex):
        return GenerativeCompletion(generate.regex, regex)

    @staticmethod
    def choices_completion(choices):
        return GenerativeCompletion(generate.choice, choices)

    @staticmethod
    def type_completion(type_name):
        if type_name not in ["float", "integer"]:
            raise Exception("type must be float or integer")
        return GenerativeCompletion(getattr(generate, type_name))

    @staticmethod
    def response_completion():
        return GenerativeCompletion(lambda _: (lambda x: x['response']))


class GrammarCompletion(Completion):
    def __init__(self, model_uri, tokenizer, grammar_type='python'):
        # Initialize the GrammarCompletion class with a model URI and tokenizer
        self.model_identifier = model_uri
        self.tokenizer = tokenizer

        # Load model from HuggingFace
        self.model = AutoModelForCausalLM.from_pretrained(self.model_identifier)
        print(f"{self.model_identifier} model is ready for use on {self.model.device}", flush=True)
        
        self.constraint = GrammarConstraint.create(grammar_type, model_uri, tokenizer)

        # Initialize dict to store terminal symbols and their corresponding token mappings
        self.terminals_tokens_map = {}

    @property
    def eos_token_id(self):
        return self.model.config.eos_token_id
    
    def complete(self, prompt, **model_kwargs):
        # Check if the specified grammar type is supported

        # Encode the input prompt
        token_ids = self.tokenizer(prompt, return_tensors='pt')
        input_len = len(token_ids['input_ids'][-1])

        # Construct terminal tokens mapping if it's not yet initialized
        if len(self.terminals_tokens_map) == 0:
            terminal_regex_map = self.constraint.parse_grammar()
            for k, v in terminal_regex_map.items():
                matching_tokens = self.constraint.construct_filter_set(v)
                self.terminals_tokens_map[k] = matching_tokens

            # Add end-of-sequence token mapping
            self.terminals_tokens_map['$END'] = [(self.tokenizer.decode(self.eos_token_id), self.eos_token_id)]

        # Initialize the LogitsProcessorList for processing sequence logits and append GrammarLogitsProcessor instance
        logits_processor = LogitsProcessorList()
        logits_processor.append(GrammarLogitsProcessor(self.constraint, self.terminals_tokens_map))

        model_kwargs["do_sample"] = model_kwargs["temperature"] > 0
        model_kwargs["output_scores"] = True
        model_kwargs["return_dict_in_generate"] = True
        model_kwargs["logits_processor"] = logits_processor

        # Generate completion using the model
        res = self.model.generate(**model_kwargs, input_ids=token_ids['input_ids'], attention_mask=token_ids['attention_mask'], eos_token_id=self.eos_token_id, pad_token_id=self.eos_token_id)

        # Decode completion and return the generated text, including input tokens
        res_text = self.tokenizer.batch_decode(res.sequences, skip_special_tokens=True)[0]
        return res_text.strip()
    

class GrammarConstraint(ABC):
    # Initialize the GrammarConstraint class with the model URI and tokenizer
    def __init__(self, model_uri, tokenizer):
        self.model_uri = model_uri
        self.tokenizer = tokenizer
        self.parser = None
        self._parser_state = None
        self._copy_state = False

    @abstractmethod
    def construct_filter_set(self, expression):
        pass

    @abstractmethod
    def parse_grammar(self):
        pass

    @staticmethod
    def create(grammar_type, model_uri, tokenizer):
        grammar_type = grammar_type.lower()
        grammars = {
            'python': PythonConstraint,
            'json': JSONConstraint,
        }

        if grammar_type not in grammars.keys():
            raise ValueError(f'{grammar_type} is not supported. The only valid grammar types are {grammars.keys()}')

        return grammars[grammar_type](model_uri, tokenizer)


class PythonConstraint(GrammarConstraint): 

    # Construct a filter set for valid tokens based on a given expression
    def construct_filter_set(self, expression):
        vocab = self.tokenizer.vocab
        valid_tokens = []
        specials = ['*', '+', ')']

        # Preprocess expression to handle special cases
        if expression[0] in specials:
            expression = f'\{expression}'
        elif len(expression) == 1 and (expression == '[' or  expression == '('):
            expression = f'\{expression}'
        
        try:
            # Compile the regex pattern and use it to match valid tokens in the vocabulary
            pattern = re.compile(expression, re.UNICODE)
            for token, id in vocab.items():
                if pattern.match(token) is not None:
                   valid_tokens.append((token, id))
        except Exception as e: 
            print("Regex Compiling Error: ", f"{e} - {expression}") 
    
        # Return a set of valid tokens based on the pattern
        return set(valid_tokens)
    
    def parse_grammar(self):
        # Create the Python parser with Lark, using the LALR algorithm
        self.parser = Lark.open_from_package('lark', 'python.lark', ['grammars'], parser='lalr', regex=True, lexer='contextual', postlex=PythonIndenter(), start='file_input')
        terminals = self.parser.terminals
        t_map = {}
        for t in terminals:
            t_map[t.name] = t.pattern.value

        # Return a map of terminal tokens and their corresponding regex patterns
        return t_map

    def _prefix_state(self, prefix_str=None, last_token=None):
        valid_next = []
        if self._parser_state is None:
            try:      
                # Parse the entire token sequence
                interactive_tree = self.parser.parse_interactive(prefix_str)
                interactive_tree.exhaust_lexer()
                self._parser_state = interactive_tree.copy()
                
                # Get the valid next states
                valid_next = list(interactive_tree.accepts())
            except Exception as e:
                # print("Parsing Error: ", e)
                pass
        else:
            try:
                # lex the last token
                last_lex = list(self.parser.lex(last_token))
                # update parser state by feeding last token
                self._parser_state.feed_token(last_lex[0])
                valid_next = list(self._parser_state.accepts())
            except Exception as e:
                # print("Parsing Error: ", e)
                pass

        # Return a list of valid next terminals
        return valid_next

    def construct_final_filter_set(self, prefix_ids, terminals_map):
        valid_next_ids = []
        
        if self._copy_state == True:
            # Decode only the last prefix ID
            last_token = self.tokenizer.batch_decode(prefix_ids[:, -1], skip_special_tokens=True)[0]
            # Get valid next terminals from the last token
            next_lex = self._prefix_state(last_token=last_token)
        else:
            # Decode all the prefix IDs
            prefix_str = self.tokenizer.batch_decode(prefix_ids, skip_special_tokens=True)[0]
            # Get valid next terminals for the prefix
            next_lex = self._prefix_state(prefix_str=prefix_str)
            self._copy_state = True

        for lex in next_lex:
            # Get the token set for each terminal symbol
            token_set = terminals_map[lex]
            for t in token_set:
                # Add valid token IDs to the list
                valid_next_ids.append(t[-1])
        
        # Return a set of valid next token IDs
        return set(valid_next_ids)


class JSONConstraint(GrammarConstraint): 

    # Construct a filter set for valid tokens based on a given expression
    def construct_filter_set(self, expression):
        vocab = self.tokenizer.vocab
        valid_tokens = []
        specials = ['*', '+', ')']

        # Preprocess expression to handle special cases
        if expression[0] in specials:
            expression = f'\{expression}'
        elif len(expression) == 1 and (expression == '[' or  expression == '('):
            expression = f'\{expression}'

        try:
            # Compile the regex pattern and use it to match all valid tokens in the vocabulary
            pattern = re.compile(expression, re.UNICODE)
            for token, id in vocab.items():
                if pattern.match(token) is not None:
                   valid_tokens.append((token, id))
        except Exception as e: 
            print(e, expression) 
    
        # Return a set of valid tokens based on the pattern
        return set(valid_tokens)
    
    def parse_grammar(self):
        # Define JSON grammar
        json_grammar = r"""
        ?start: value

        ?value: object
            | array
            | string
            | SIGNED_NUMBER      -> number
            | "true"             -> true
            | "false"            -> false
            | "null"             -> null

        array  : "[" [value ("," value)*] "]"
        object : "{" [pair ("," pair)*] "}"
        pair   : string ":" value

        string : ESCAPED_STRING

        %import common.ESCAPED_STRING
        %import common.SIGNED_NUMBER
        %import common.WS

        %ignore WS
        """
        
        # Create Lark internal transformer to make parsing faster and more memory efficient 
        class TreeToJson(Transformer):
            @v_args(inline=True)
            def string(self, s):
                return s[1:-1].replace('\\"', '"')

            array = list
            pair = tuple
            object = dict
            number = v_args(inline=True)(float)

            null = lambda self, _: None
            true = lambda self, _: True
            false = lambda self, _: False


        # Create the JSON parser with Lark, using the LALR algorithm
        self.parser = Lark(json_grammar, parser='lalr',
                        lexer='contextual',
                        transformer=TreeToJson())
        terminals = self.parser.terminals
        t_map = {}
        for t in terminals:
            t_map[t.name] = t.pattern.value
        
        # Return a map of terminal tokens and their corresponding regex patterns
        return t_map

    def _prefix_state(self, prefix_str=None, last_token=None):
        valid_next = []
        if self._parser_state is None:
            try:      
                # Parse the entire token sequence
                interactive_tree = self.parser.parse_interactive(prefix_str)
                interactive_tree.exhaust_lexer()
                self._parser_state = interactive_tree.copy()
                
                # Get the valid next states
                valid_next = list(interactive_tree.accepts())
            except Exception as e:
                # print("Parsing Error: ", e)
                pass
        else:
            try:
                # lex the last token
                last_lex = list(self.parser.lex(last_token))
                # update parser state by feeding last token
                self._parser_state.feed_token(last_lex[0])
                valid_next = list(self._parser_state.accepts())
            except Exception as e:
                # print("Parsing Error: ", e)
                pass

        # Return a list of valid next terminals
        return valid_next

    def construct_final_filter_set(self, prefix_ids, terminals_map):
        valid_next_ids = []
        
        if self._copy_state == True:
            # Decode only the last prefix ID
            last_token = self.tokenizer.batch_decode(prefix_ids[:, -1], skip_special_tokens=True)[0]
            # Get valid next terminals from the last token
            next_lex = self._prefix_state(last_token=last_token)
        else:
            # Decode all the prefix IDs
            prefix_str = self.tokenizer.batch_decode(prefix_ids, skip_special_tokens=True)[0]
            # Get valid next terminals for the prefix
            next_lex = self._prefix_state(prefix_str=prefix_str)
            self._copy_state = True

        for lex in next_lex:
            # Get the token set for each terminal symbol
            token_set = terminals_map[lex]
            for t in token_set:
                # Add valid token IDs to the list
                valid_next_ids.append(t[-1])
                
        # Return a set of valid next token IDs
        return set(valid_next_ids)


class GrammarLogitsProcessor(LogitsProcessor):

    def __init__(self, constraint_class, terminals_map):
        # Initialize the GrammarLogitsProcessor class with a constraint class and terminals map.
        self.constraint_class = constraint_class 
        self.terminals_map = terminals_map  

    def __call__(self, input_ids, scores):
        # This method is called for each generation step
        
        # Initialize a boolean bias tensor with the same shape as scores
        bias = torch.zeros_like(scores, dtype=torch.bool)
        
        # Get the set of valid next token IDs for current step
        valid_next_ids = self.constraint_class.construct_final_filter_set(input_ids, self.terminals_map)
        
        # Set the bias to True for valid next token IDs
        for id in valid_next_ids:
            bias[0, id] = True
        
        # Add the bias to the scores tensor to zero-out invalid next tokens
        scores += bias
        
        # Return the modified scores tensor
        return scores



import re
import os
import argparse
import sys
from dynaconf import Dynaconf
from xdg import XDG_CONFIG_HOME
from llm_vm.onsite_llm import model_keys_registered
from llm_vm.data_path import project_root
from ipaddress import ip_address


# project_root = os.path.abspath(os.getcwd())

print("Project Root: " + project_root, file=sys.stderr)


# default to the local file, and look in XDG if we can't find it
config_files = [
    os.path.join(project_root, "settings.default.toml"),
    os.path.join(XDG_CONFIG_HOME, "settings.toml"),
    os.path.join(project_root, "settings.toml"),
]

settings = Dynaconf(
    settings_file=config_files,
    load_dotenv=True,
    dotenv_path=".env",
    envvar_prefix="LLM_VM",
)

# making MODELS_AVAILABLE a set because it will be used for membership testing
MODELS_AVAILABLE = set(model_keys_registered)

if settings.big_model not in MODELS_AVAILABLE:
    print(settings.big_model + " is an invalid Model selection for Big LLM Model", file=sys.stderr)
    exit()

if settings.small_model not in MODELS_AVAILABLE:
    print(settings.small_model + " is an invalid Model selection for Small LLM Model", file=sys.stderr)
    exit()

# do we want to do this early test and exit?
# if settings.small_model is "chat_gpt":
#     print("openai currently doesn't support fine tuning chat_gpt, aka gpt3.5-turbo, exiting")
#     exit()

def isOpenAIModel(str):
    return str =="gpt" or str =="chat_gpt"

if settings.openai_api_key is None and (isOpenAIModel(settings.small_model) or isOpenAIModel(settings.big_model)):
    print("Error: you must have an OpenAI API key set via config files, ./settings.default.toml or via environment variable ", file=sys.stderr)
    print("LLM_VM_OPEN_AI_API,if you wish to use their models. Exiting", file=sys.stderr)
    exit()

# check settings.host is a valid IP address
def isValidIP(ip):
    try:
        # if object created IP is valid
        tmp = ip_address(ip)
        return True
    except ValueError:
        return False

if isValidIP(settings.host) is False:
    print("Invalid IP4/IP6 address. Reverting to the default host: 127.0.0.1", file=sys.stderr)
    settings.host = '127.0.0.1'



assert settings.big_model in MODELS_AVAILABLE, f"{settings.big_model} is not a valid Model selection for Big LLM Model"
assert settings.small_model in MODELS_AVAILABLE, f"{settings.small_model} is not a valid Model selection for Small LLM Model"
assert 1024 < settings.port < 65535, f"{settings.port} is an invalid port number"



"""
This file has been temporarily repurposed as an interface for users to
call any of the three agents (REBEL, BACKWARD_CHAINING, and FLAT) and interact with them.
Running this file prompts the user to choose any of the agents and ask it questions.
"""
import sys
import llm_vm.agents.REBEL.agent as REBEL
import llm_vm.agents.FLAT.agent as FLAT
import os
key = os.getenv("LLM_VM_OPENAI_API_KEY")

def call_agent():
    print("Try out any agent!", file=sys.stderr)

    # stores user input for which agent to try out
    model_choice = 0

    # times that the user was prompted to choose a valid model
    times_asked_for_model = 0

    # if user enters invalid choice, prompt for input until valid
    while True:
        model_choice = input("[1] FLAT\n[2] REBEL\nChoose your agent:")
        try:
            # try to cast the input to an integer
            model_choice = int(model_choice)

            if model_choice not in range (1, 4):
                print("=====Please enter 1, or 2!=====", file=sys.stderr)
            else:
                # user has entered a valid input
                break
        except:
            print("=====Please enter 1 or 2!=====", file=sys.stderr)

    # FLAT
    if model_choice == 1:
        # TODO: Add agent call here when FLAT is fixed
        tools =  [{'method': 'GET', "dynamic_params": { 'location': 'This string indicates the geographic area to be used when searching for businesses. \
    Examples: "New York City", "NYC", "350 5th Ave, New York, NY 10118".', 'term': 'Search term, e.g. "food" or "restaurants". The \
    term may also be the business\'s name, such as "Starbucks"', 'price': 'Pricing levels to filter the search result with: 1 = \
    $, 2 = $$, 3 = $$$, 4 = $$$$. The price filter can be a list of comma delimited pricing levels. e.g., "1, 2, 3" will filter the \
    results to show the ones that are $, $$, or $$$.'}, "description":"This tool searches for a business on yelp.  It's useful for finding restaurants and \
    whatnot.", 'args' :{'url': 'https://api.yelp.com/v3/businesses/search', 'cert': '', 'json': {}, 'params': {'limit': '1',
                                                                                                              'open_now': 'true', 'location': '{location}', 'term': '{term}', 'price': '{price}'}, 'data': {},
                       'headers': {'authorization': '',
                                   'accept': 'application/json'}}}]
        agent = FLAT.Agent(key, tools, verbose=1)
    elif model_choice == 2:
        tools = REBEL.buildExampleTools()
        agent = REBEL.Agent(key, tools, verbose = 1)


    pass

    mem = []
    last = ""
    while True:
        inp = input(last+"Human: ")
        ret = agent.run(inp, mem)
        mem = ret[1]
        last = "AI: "+str(ret[0])+ "\n"

if __name__ == "__main__":
    call_agent()



"""
This file contains common tools (JSON API request objects) that are used
for all the agents to retrieve information for questions across various topics.
"""

import os
import sys

# Get the current file's directory to grab the python files with common functionality in the utils/ folder
current_dir = os.path.dirname(os.path.abspath(__file__))
grandparent_dir = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
utils_dir = os.path.join(grandparent_dir, 'utils/')
sys.path.append(utils_dir)

from llm_vm.utils.keys import *
from llm_vm.utils.labels import *
from llm_vm.utils.typings_llm import *

CUSTOM_TOOL_ANSWER_EMBEDDING = "/answer_embedding"


def __get_example_tools():
    # wolfram
    google_tool = {
               'description': "The tool returns the results of free-form queries similar to those used for wolfram alpha. This is useful for complicated math or live data retrieval.  Can be used to get the current date.",
               'dynamic_params': {"q": 'The natural language input query'},
               'method': 'GET',
               'args': {'url': "https://www.googleapis.com/customsearch/v1",
                         'params': {'key': "",
                                    'cx' : "",
                                    'q': '{q}'}
                        }
               }

    # geopy
    directions_tool = {'description': "Find the driving distance and time to travel between two cities.",
               'dynamic_params': {"origins": 'the origin city', "destinations": 'the destination city'},
               'method': 'GET',
               'args': {'url': "https://maps.googleapis.com/maps/api/distancematrix/json",
                         'params': {'key': "",
                                    'origins': '{origins}',
                                    'destinations': '{destinations}'}
                        }}

    # weather
    weather_tool = {'description': 'Find the weather at a location and returns it in celcius.',
               'dynamic_params': {"latitude": 'latitude of as a float',
                                  "longitude": 'the longitude as a float'},
               'method': 'GET',
               'args': {'url': "https://api.open-meteo.com/v1/forecast",
                         'params': {'current_weather': 'true',
                                    'latitude': '{latitude}',
                                    'longitude': '{longitude}'
                                    }}
               }
    return [google_tool, directions_tool, weather_tool]


GENERIC_TOOLS = __get_example_tools()



import flask
from flask import request, jsonify
import json
import time
import traceback
import importlib
import openai
import os
import hashlib
import sys
from flask_cors import CORS
from contextlib import contextmanager
import llm_vm.server.routes as routes
from llm_vm.config import settings
from llm_vm.utils.ram import RAMLogger

app = flask.Flask(__name__)
CORS(app)
app.config["DEBUG"] = True

# Register_blueprint from routes to load API
app.register_blueprint(routes.bp)

def server_entry_point(host = '127.0.0.1', port = 3002):
    """
    This function launches the server with specified parameters

     Parameters:
         host (str): Network IP address
         port (int): Port Number

     Returns:
         None

     Example:
         >>> server_entry_point(port = 3002)
    """
    app.run(host = host,port = port)

def cli():
    """
     This function is the entry point for the project and allows the user to specify an option network address and port number when launching from the cli

     Parameters:
         None

     Returns:
         None

    """
    logger = RAMLogger()
    logger.start()
    port = settings.port
    if port > 65535:
        print('Port defined out of range, defaulting to 3002', file=sys.stderr)
        port = 3002
    try:
        server_entry_point(host = settings.host, port = port)
    finally:
        logger.end()

if __name__ == '__main__':
    cli()



import os
import sys

import requests
from llama_index import Document, GPTTreeIndex
from llm_vm.agents.FLAT.agent_helper.utils import print_op, make_interaction, print_big
from llm_vm.agents.FLAT.agent_helper.requests.call_llm import call_llm
from llm_vm.agents.FLAT.agent_helper.replacer import replace_variables_for_values
from llm_vm.utils.typings_llm import *
from llm_vm.agents.FLAT.agent_helper.bothandler import prompt_for_answer, prompt_for_instructions

# # Get the current file's directory to grab the python files with common functionality in the utils/ folder
# current_dir = os.path.dirname(os.path.abspath(__file__))
# grandparent_dir = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
# utils_dir = os.path.join(grandparent_dir, 'utils/')
# sys.path.append(utils_dir)

from llm_vm.utils.labels import *
from llm_vm.agents.FLAT.agent_helper.tools import CUSTOM_TOOL_ANSWER_EMBEDDING
import json

def __format_tool_url (url: str) -> str:
    if CUSTOM_TOOL_ANSWER_EMBEDDING in url:
        return "CHAT.DEV Embeddings Service"
    return url


def use_tool(
    tool: SingleTool,
    gpt_suggested_input: dict,
    question: str,
    memory: TupleList,
    verbose: int,
    bot_instructions: str,
    response_instructions: str = ""
) -> Tuple[str, str, Any, float]:
    price = 0

    # make sure all of the suggested fields exist in the tool desc
    for i in gpt_suggested_input.keys():
        if i not in tool["dynamic_params"].keys():
            raise Exception("Bad Generated Input")

    try:
        tool_args = replace_variables_for_values(
            tool["args"],
            gpt_suggested_input
        )
    except Exception as e:
        print_big(e, "EXCEPTION");
        tool_args = {}

    url = tool_args['url']

    if 'auth' in tool_args and isinstance(tool_args['auth'], dict):
        auths = list(tool_args['auth'].items())
        if len(auths) > 0:
            tool_args['auth'] = list(tool_args['auth'].items())[0]
        else:
            del tool_args['auth']

    # Remove those parameters that are not part of Python's `requests` function.
    tool_args.pop("jsonParams", None)
    tool_args.pop("urlParams", None)

    # print_big(tool_args, "ARGS")
    request_function = None
    method = tool['method'].upper()
    if method == "PUT":
        request_function = requests.put
    elif method == "POST":
        request_function = requests.post
    elif method == "PATCH":
        request_function = requests.patch
    elif method == "DELETE":
        request_function = requests.delete
    elif method == "GET":
        request_function = requests.get
    else:
        request_function = requests.get

    resp = request_function(**tool_args)
    print_big(resp.url, "FINAL URL: (" + tool["method"] + ") ")

    api_call_args = str(tool_args)
    return_value = str(resp.text)

    formatted_url = __format_tool_url(str(url))

    if verbose > -1:
        pass
        #commented out, causes test cases to fail because the return json of some wolfram alpha searches contains character encodings that charmap does not know.
        # leads to error being thrown.
        # print_op("URL Response:", ret)

    try:
        # seems pointless, but it formats the resulting json without spaces.
        # If the response isn't json, it does nothing.
        return_value_json = json.loads(return_value)
        return_value = str(return_value_json)
    except:
        return_value_json = {"response": return_value}
        pass

    if resp.status_code in [400, 401, 404, 500]:
        er = " → " + str(resp.status_code)
        return "This tool isn't working currently" + er, formatted_url + er, return_value_json, price


    summarised_ret: str = ""
    if len(return_value) > 8000:
        document = Document(return_value)
        try:
            if verbose > 2:
                print_op("### Summarising with GPT Tree Index ...")
            summarised_ret = GPTTreeIndex([document]).query(
                response_instructions, response_mode="tree_summarize")
        except:
            return "OpenAI is down!", formatted_url, None, price

        if verbose > 2:
            print_op("SUMMARISED:", summarised_ret)

        if str(summarised_ret).find("ANSWER: ") == 0:
            summarised_ret = str(return_value)[len("ANSWER: X."):].strip(" .")

    else:

        mem = "".join([make_interaction(p, a, INTERACTION = L_INT_HUMAN) for p,a in memory]) \

        # THIS WORKS VERY WELL, but using <XML> tags here does not work.
        summarise_prompt, stop = f"{bot_instructions}<{L_CONVERSATION}>{mem}</{L_CONVERSATION}>\n\
Description: a JSON object contains:\n{return_value}\nUsing this information, what is the [{L_ANSWER}] to [{L_QUESTION}]?\n \
{prompt_for_instructions(response_instructions)}{prompt_for_answer(question)}", None

        if verbose > 3:
            print_big(summarise_prompt, "SUMMARY OF JSON DECODE")

        summarised_ret, price = call_llm({
            "llm": LLMCallType.OPENAI_COMPLETION,
            "model": OpenAIModel.DAVINCI_TEXT.value,
            "max_tokens": 256,
            "stop": stop,
            "prompt": summarise_prompt,
            "temperature": 0.1,
        })
        print_big(summarised_ret, "SUMMARISED ANSWER")

    return summarised_ret, formatted_url, return_value_json, price



import re
from llm_vm.utils.keys import DICT_KEY_REGEX_TO_FIND_PURE_INTERPOLATIONS

###
# Use this regex to find if a variable is a pure interpolation.
# For @example, { "name": "{my_name}" } is pure, since there's nothing else besides the interpolation
# For @example, { "Ticker": "TickerID={ticker_id}|historical=1" } is NOT pure, since there is extra text
def __is_pure_interpolation (key: str) -> bool:
    return len(re.findall(DICT_KEY_REGEX_TO_FIND_PURE_INTERPOLATIONS, key)) == 1

###
# Use this to get the key of a pure interpolation ===> '{a}' => 'a', '{abc_def-345}' => 'abc_def-345'
def __get_pure_key (key: str) -> str:
    return key[1:-1]


def replace_variables_for_values(my_dict: dict, dynamic_keys: dict, ignore_key: str = "_______") -> any:
    def format_simple_value(value: str) -> any:
        try:
            if __is_pure_interpolation(value):
                formatted_value = dynamic_keys[__get_pure_key(value)]
            else:
                formatted_value = value.format(**dynamic_keys)
        except Exception as e:
            formatted_value = value

        return formatted_value

    replaced_dict = {}
    for key, value in my_dict.items():
        if (key == ignore_key):
            continue
        formatted_key = key.format(**dynamic_keys)
        if (isinstance(value, dict)):
            if (not isinstance(value, list) and not isinstance(value, dict)):
                formatted_value = format_simple_value(value)
            else:
                formatted_value = replace_variables_for_values(value, dynamic_keys, ignore_key)
        elif (isinstance(value, list)):
            formatted_value = []
            for item in value:
                if (not isinstance(item, list) and not isinstance(item, dict)):
                    formatted_value.append(format_simple_value(item))
                else:
                    formatted_value.append(replace_variables_for_values(item, dynamic_keys, ignore_key))
        else:
            formatted_value = format_simple_value(value)
        replaced_dict[formatted_key] = formatted_value
    return replaced_dict



import os
import sys
import re
import random
from llm_vm.agents.FLAT.agent_helper.business_logic import promptf
from llm_vm.utils.labels import *
from llm_vm.agents.FLAT.agent_helper.utils import *
from llm_vm.agents.FLAT.agent_helper.tools import *
from llm_vm.utils.keys import *


random_fixed_seed = random.Random(4)

class Agent:
    def __init__(self, openai_key, tools = None, bot_instructions = "", verbose = 4):
        self.verbose = verbose
        self.set_tools((GENERIC_TOOLS + tools) if tools else GENERIC_TOOLS)
        self.bot_instructions = f"<{L_BOT_INSTRUCTIONS}>{bot_instructions}<{L_BOT_INSTRUCTIONS}>" if bot_instructions else ""

        # set the openai key to make calls to the API
        set_api_key(openai_key)
    def set_tools(self, tools):
        self.tools = []
        for tool in tools:
            if not 'args' in tool:
                tool['args'] = {}
            if not 'method' in tool:
                tool['method'] = "GET"
            if not 'examples' in tool:
                tool['examples'] = []
            if not 'dynamic_params' in tool:
                tool['dynamic_params'] = {}
            if not 'id' in tool:
                tool['id'] = len(self.tools) + 1

            self.tools += [tool]


    def run(self, question: str, memory: TupleList) -> Tuple[str, TupleList, list, DebugCallList]:
        try:
            answer, _, calls, debug_return, price = promptf(
                question,
                memory,
                self.tools,
                self.bot_instructions,
                self.verbose
            )

            # Remove xml tags from answer, if any. Important, because this is trade secret.
            answer, has_friendly_tags = remove_tags_from_html_string(answer)

        except Exception as e:
            print(e, flush=True, file=sys.stderr)
            print("Main thread exception: ", e, flush=True, file=sys.stderr)
            answer, calls, debug_return, price, has_friendly_tags = "Error: " + str(e), [], [], 0, False


        if self.verbose > -1:
            print_big("GPT-3.5 Price = ~{:.1f} cents".format(price * 100))

        return (answer, memory + [(question, answer)], calls, debug_return, has_friendly_tags)

def flat_main():
    # tools = [{'method': 'GET',"description":"use this tool to find the price of stocks",'args' : {"url":"https://finnhub.io/api/v1/quote",'params': { 'token' :'cfi1v29r01qq9nt1nu4gcfi1v29r01qq9nt1nu50'} },"dynamic_params":{"symbol":"the symbol of the stock"}}]
    tools =  [{'method': 'GET', "dynamic_params": { 'location': 'This string indicates the geographic area to be used when searching for businesses. \
    Examples: "New York City", "NYC", "350 5th Ave, New York, NY 10118".', 'term': 'Search term, e.g. "food" or "restaurants". The \
    term may also be the business\'s name, such as "Starbucks"', 'price': 'Pricing levels to filter the search result with: 1 = \
    $, 2 = $$, 3 = $$$, 4 = $$$$. The price filter can be a list of comma delimited pricing levels. e.g., "1, 2, 3" will filter the \
    results to show the ones that are $, $$, or $$$.'}, "description":"This tool searches for a business on yelp.  It's useful for finding restaurants and \
    whatnot.", 'args' :{'url': 'https://api.yelp.com/v3/businesses/search', 'cert': '', 'json': {}, 'params': {'limit': '1',
                                                                                                              'open_now': 'true', 'location': '{location}', 'term': '{term}', 'price': '{price}'}, 'data': {},
                       'headers': {'authorization': 'Bearer OaEqVSw9OV6llVnvh9IJo92ZCnseQ9tftnUUVwjYXTNzPxDjxRafYkz99oJKI9WHEwUYkiwULXjoBcLJm7JhHj479Xqv6C0lKVXS7N91ni-nRWpGomaPkZ6Z1T0GZHYx',
                                   'accept': 'application/json'}}}]

    label = Agent(os.getenv("LLM_VM_OPENAI_API_KEY"), tools, verbose=4)
    conversation_history = []
    last = ""
    while True:
        inp = input(last+"Human: ")
        return_value = label.run(inp, conversation_history)
        conversation_history = return_value[1]
        print(return_value[2], file=sys.stderr)
        last = "AI: "+str(return_value[0]) + "\n"

# print_op(google(' {"question": ""}'))
if __name__ == "__main__":
    flat_main()



import openai
import os
from llm_vm.utils.typings_llm import *

def __tokens_to_dollars(usage) -> float:
    return float(usage)/1000*0.02

def call_open_ai(request: LLMCallParams) -> LLMCallReturnType:
    model = request['model']
    max_tokens = request['max_tokens']
    stop = request['stop'] if 'stop' in request else None
    cur_prompt = request['prompt']
    temperature = request['temperature'] if 'temperature' in request else 0.0
    chat = request['llm']

    if isinstance(model, tuple):
        model, api_key = model[0], model[1]
    else:
        api_key = False

    if api_key:
        if api_key[0] == '$':
            api_key = False


    current_key = None
    if api_key:
        current_key = openai.api_key
        os.environ["OPENAI_API_KEY"] = api_key
        openai.api_key = api_key
    else:
        api_key = os.getenv("LLM_VM_OPENAI_API_KEY")
        openai.api_key = api_key

    if chat == LLMCallType.OPENAI_CHAT:
        response = openai.ChatCompletion.create(
                model="gpt-3.5-turbo-0301",
                max_tokens=max_tokens,
                stop=stop,
                messages=cur_prompt,
                temperature=temperature
            )
        response_text = response['choices'][0]['message']['content']
    elif chat == LLMCallType.OPENAI_COMPLETION:

        response = openai.Completion.create(
                model=model,
                max_tokens=max_tokens,
                stop=stop,
                prompt=cur_prompt,
                temperature=temperature
            )

        response_text = response['choices'][0]['text']

    if current_key:
        openai.api_key = current_key
        os.environ["OPENAI_API_KEY"] = current_key

    usage = response["usage"]["total_tokens"]
    price = __tokens_to_dollars(usage)
    return response_text.strip("\n "), price



from llm_vm.utils.typings_llm import *
import os
import json
from llm_vm.agents.FLAT.agent_helper.utils import verbose_answer

def __get_tool_input_model(model: OpenAIModel) -> ToolInputModel:
    current_dir = os.path.dirname(os.path.abspath(__file__))
    file_path = os.path.abspath(os.path.join(current_dir, f'../../raw_data/tool_input_data.json'))

    json_data: List[ToolInputModelJSONData] = json.load(open(file_path, "r"))["data"]

    def __get_mem_tuple(json_mem: List[List[Union[str, dict]]]) -> TupleList:
        mem: TupleList = []
        for m in json_mem:
            if len(m) == 3:
                # when the "mem" has also the previous JSON response
                mem.append((m[0], verbose_answer(m[2], m[1])))
            elif len(m) == 2:
                # when the "mem" has only question/answer
                mem.append((m[0], m[1]))
        return mem

    return {
        "openai_model": model,
        "data": [{
            "mem": __get_mem_tuple(t["mem"]) if "mem" in t else [],
            "question": t["question"],
            "answer": t["answer"],
            "description": t["description"],
            "params": t["params"]
        } for t in json_data]
    }


tool_input_data: ToolInputModel = __get_tool_input_model(OpenAIModel.DAVINCI)



from llm_vm.agents.FLAT.agent_helper.tool_utils import splitter_prompt
from llm_vm.agents.FLAT.agent_helper.labels import *
from llm_vm.typings_llm import *
from llm_vm.agents.FLAT.models.utils.question_split_model.question_split_model_data import question_splitter_data

def __construct_question_split_jsonl(data: List[QuestionSplitModelData]) -> List[PromptModelEntry]:
    jsonl_entries: List[PromptModelEntry] = []
    for entry in data:
        prompt, stopper = splitter_prompt([{
            "mem": entry["mem"] if "mem" in entry else None,
            "tools": entry["tools"] if "tools" in entry else None,
            "question": entry["question"]
        }])
        jsonl_entries.append({
            "prompt": prompt,
            "completion": SUBQ_SPLITTER.join(entry["answer"]) + stopper
        })
    return jsonl_entries


def question_splitter_jsonl() -> Tuple[PromptModelEntry, str]:
    return (
        __construct_question_split_jsonl(question_splitter_data["data"]),
        question_splitter_data["openai_model"].value
    )



import sys
import openai
import llm_vm.onsite_llm as llms
from llm_vm.onsite_llm import load_model_closure
from llm_vm.agents.REBEL import agent
from llm_vm.completion.optimize import LocalOptimizer
import llm_vm.config as conf
from llm_vm.vector_db import PineconeDB
import os
import PyPDF2
from sentence_transformers import SentenceTransformer


if conf.settings.big_model is None:
    default_big_model = "chat_gpt"
else:
    default_big_model= conf.settings.big_model

if conf.settings.small_model is None:
    default_small_model= "pythia"
else:
    default_small_model = conf.settings.small_model

# Builds the client to facilitate easy use.
def client_build(type = "inference", openai_key = None, big_model = default_big_model, small_model =default_small_model,big_model_config={},small_model_config={}):
    if type == "inference":
        return Simple_Inference_Client(big_model, big_model_config, openai_key)
    else:
        return Client(big_model, small_model, big_model_config, small_model_config)
    
# Shoe-in for simple inference. Allows for traditional initialization of the client or dynamic initialization for rapid inference without agents.
class Simple_Inference_Client:
    def __init__(self, model = default_big_model, model_config = {}, openai_key = None):
        self.model = load_model_closure(model)(**model_config)
        openai.api_key = openai_key

    def complete(self, prompt, max_len=256, **kwargs):
        return self.model.generate(prompt, max_len,**kwargs)



class Client:
    """
    This is a class sets up a Local optimizer around a LLM for easy access.

    Attributes:
        optimizer (LocalOptimizer): Anarchy LLM Optimizer
        openai_key (str): API key for OpenAI access

    Methods:
        complete: The Anarchy completion endpoint giving access to a specified LLM
    """

    def __init__(self, big_model = default_big_model, small_model =default_small_model,big_model_config={},small_model_config={}):
        """
        This __init__ function allows the user to specify which LLM they want to use upon instantiation.

        Parameters:
            big_model (str): Name of LLM to be used as a reliable source
            small_model (str): Name of small model to be used for fine-tuning

        Returns:
            None

        Example:
            >>> client = Client(big_model = 'neo')
        """
        self.big_model = big_model
        self.small_model = small_model 
        self.teacher = load_model_closure(big_model)(**big_model_config)
        self.student = load_model_closure(small_model)(**small_model_config)
        self.vector_db = None
        self.emb_model = None

        ## FIXME, do something like $HOME/.llm_vm/finetuned_models/
        if os.path.isdir("finetuned_models") == False:
            os.mkdir("finetuned_models")
        # Specify the model strategy the user will use
        # MODELCONFIG = models.ModelConfig(big_model=big_model, small_model=small_model)
        print("Using model: " + big_model, file=sys.stderr) # announce the primary LLM that is generating results

        # These functions allow for proper initialization of the optimizer
        # def CALL_BIG(prompt, max_len=256, **kwargs):

        #     return self.teacher.generate(prompt, max_len,**kwargs)

        def CALL_SMALL(prompt, max_len=256, **kwargs):
            small_model_filename = kwargs.get('small_model_filename', None)
            if small_model_filename is not None:
                self.student.load_finetune(small_model_filename)

            return self.student.generate(prompt, max_len,**kwargs)

        # load the optimizer into object memory for use by the complete function
        self.optimizer = LocalOptimizer(MIN_TRAIN_EXS=2,openai_key=None, call_big=self.CALL_BIG, call_small= CALL_SMALL,
                                        big_model = self.teacher, small_model = self.student)
        self.rebel_agent = None # only initialized on first use

    # These functions allow for proper initialization of the optimizer
    def CALL_BIG(self, prompt, max_len=256, **kwargs):

        return self.teacher.generate(prompt, max_len,**kwargs)

    def complete(self, prompt,
                 context = "", # TODO: change back to static_context because otherwise 
                 openai_key = None,
                 finetune=False,
                 data_synthesis = False,
                 temperature=0,
                 stoptoken = None,
                 tools = None,
                 openai_kwargs = {},
                 hf_kwargs = {}):
        """
        This function is Anarchy's completion entry point

        Parameters:
            prompt (str): Prompt to send to LLM for generation
            context (str): Unchanging context to send to the LLM for generation.  Defaults to "" and doesn't do fine-tuning.
            finetune (bool): Boolean value that begins fine tuning when set to True
            data_synthesis (bool): Boolean value to determine whether data should be synthesized for fine-tuning or not
            temperature (float): Sampling temperature for use during generation, between 0 and 2
            stoptoken (str | list): Sequence for stopping token generation
            tools (list): API tools for using with the Rebel agent
            openai_kwargs (dict): Key word arguments for generation with Open AI models
            hf_kwargs (dict): Key word arguments for generation with 


        Returns:
            str: LLM Generated Response

        Example:
           >>> SmallLocalOpt.generate("How long does it take for an apple to grow?)
           How long does it take for an apple tree to grow?
        """
        static_context = context
        dynamic_prompt = prompt
        use_rebel_agent = False

        if self.big_model in ['gpt', 'chat_gpt', 'gpt4']:
            kwargs = openai_kwargs
            if "temperature" not in kwargs:
                kwargs.update({"temperature":temperature})
            if stoptoken is not None:
                kwargs.update({"stop":stoptoken})
                
        else:
            kwargs = hf_kwargs
            if temperature > 0:
                kwargs.update({"do_sample": True})
                kwargs.update({"temperature":temperature})

        if openai_key is not None:
            self.openai_key = openai_key

        if tools is not None:
            if type(tools) != list:
                return Exception("Wrong data type for tools. Should be a list")
            else:
                final_tools=[]
                for i in tools:
                    temp_tool_dict = {}
                    temp_args_dict = {}
                    temp_tool_dict.update({"description":i["description"]})
                    temp_tool_dict.update({"dynamic_params":i["dynamic_params"]})
                    temp_tool_dict.update({"method":i["method"]})
                    temp_args_dict.update({"url":i["url"]})
                    temp_args_dict.update({"params":{}})
                    for j in i["static_params"].keys():
                        temp_args_dict["params"].update({j:i["static_params"][j]})
                    for k in i["dynamic_params"].keys():
                        temp_args_dict["params"].update({k:"{"+k+"}"})
                    temp_tool_dict.update({"args":temp_args_dict})
                    final_tools.append(temp_tool_dict)
                if self.rebel_agent is None:
                    self.rebel_agent = agent.Agent("", [], verbose=1) # initialize only if tools registered
                self.rebel_agent.set_tools(final_tools)
                use_rebel_agent = True

        try:
            if openai_key is not None:
                openai.api_key = self.openai_key
        except:
            return  {"status":0, "resp":"Issue with OpenAI key"}

        self.optimizer.openai_key = openai.api_key
        # self.agent.set_api_key(openai.api_key,"OPENAI_API_KEY") #
        if os.getenv("OPENAI_API_KEY") is None and use_rebel_agent==True :
            print("warning: you need OPENAI_API_KEY environment variable for ", file=sys.stderr)

        try:
            if not use_rebel_agent:
                completion = self.optimizer.complete(static_context,dynamic_prompt,data_synthesis=data_synthesis,finetune = finetune, **kwargs)
                print("COMPLETED RUNNING")
            else:
                completion = self.rebel_agent.run(static_context+dynamic_prompt,[])[0]
        except Exception as e:
            return {"status":0, "resp": str(e)}
        return {"completion":completion, "status": 200}

    def load_finetune(self, model_filename=None):
        self.teacher.load_finetune(model_filename)

    def set_pinecone_db(self, api_key, env_name):
        self.vector_db = PineconeDB(api_key, env_name)
        self.emb_model = SentenceTransformer("sentence-transformers/all-MiniLM-L6-v2")

    def _pdf_loader(self, pdf_file_path):

        try:
        # Open the PDF file in read-binary mode
            with open(pdf_file_path, 'rb') as pdf_file:
                pdf_reader = PyPDF2.PdfFileReader(pdf_file)

                pdf_text = ''
                
                for page_num in range(pdf_reader.numPages):
                    page = pdf_reader.getPage(page_num)
                    pdf_text += page.extractText()
                
                pdf_file.close()
                
                return pdf_text
        
        except Exception as e:
            # Handle exceptions, such as file not found or invalid PDF format
            print(f"Error: {e}")
            return None
        
    def create_pinecone_index(self, **kwargs):
        self.vector_db.create_index(**kwargs)

    def store_pdf(self, pdf_file_path, chunk_size=1024, **kwargs):
        text = self._pdf_loader(pdf_file_path)
        chunks = [text[i:i+chunk_size] for i in range(0, len(text), chunk_size)]
        embeddings = [self.emb_model.encode(chunk) for chunk in chunks]
        kwargs["vector"] = embeddings
        self.vector_db.upsert(**kwargs)

    def RAG_complete(self, prompt,
                 context = "",
                 openai_key = None,
                 finetune=False,
                 data_synthesis = False,
                 temperature=0,
                 stoptoken = None,
                 tools = None,
                 query_kwargs = {},
                 hf_kwargs = {},
                 openai_kwargs = {}):
        
        """
        This function is Anarchy's completion entry point

        Parameters:
            prompt (str): Prompt to send to LLM for generation
            context (str): Unchanging context to send to the LLM for generation.  Defaults to "" and doesn't do fine-tuning.
            finetune (bool): Boolean value that begins fine tuning when set to True
            data_synthesis (bool): Boolean value to determine whether data should be synthesized for fine-tuning or not
            temperature (float): Sampling temperature for use during generation, between 0 and 2
            stoptoken (str | list): Sequence for stopping token generation
            tools (list): API tools for using with the Rebel agent
            query_kwargs (dict): Key word arguments for querying the Pinecone vector database
            openai_kwargs (dict): Key word arguments for generation with Open AI models
            hf_kwargs (dict): Key word arguments for generation with 

        Returns:
            str: RAG Response
        """
        
        prompt_embeddings = self.emb_model.encode(prompt)
        query_kwargs["vector"] = prompt_embeddings
        if "top_k" not in query_kwargs:
            query_kwargs["top_k"] = 5

        topk_vecs = self.vector_db.query(**query_kwargs)

        decoded_vecs = self.emb_model.decode(topk_vecs)

        prompt = decoded_vecs + "/n" + prompt

        static_context = context
        dynamic_prompt = prompt
        use_rebel_agent = False

        if self.big_model in ['gpt', 'chat_gpt', 'gpt4']:
            kwargs = openai_kwargs
            if "temperature" not in kwargs:
                kwargs.update({"temperature":temperature})
            if stoptoken is not None:
                kwargs.update({"stop":stoptoken})       
        else:
            kwargs = hf_kwargs
            if temperature > 0:
                kwargs.update({"do_sample": True})
                kwargs.update({"temperature":temperature})

        if openai_key is not None:
            self.openai_key = openai_key

        if tools is not None:
            if type(tools) != list:
                return Exception("Wrong data type for tools. Should be a list")
            else:
                final_tools=[]
                for i in tools:
                    temp_tool_dict = {}
                    temp_args_dict = {}
                    temp_tool_dict.update({"description":i["description"]})
                    temp_tool_dict.update({"dynamic_params":i["dynamic_params"]})
                    temp_tool_dict.update({"method":i["method"]})
                    temp_args_dict.update({"url":i["url"]})
                    temp_args_dict.update({"params":{}})
                    for j in i["static_params"].keys():
                        temp_args_dict["params"].update({j:i["static_params"][j]})
                    for k in i["dynamic_params"].keys():
                        temp_args_dict["params"].update({k:"{"+k+"}"})
                    temp_tool_dict.update({"args":temp_args_dict})
                    final_tools.append(temp_tool_dict)
                if self.rebel_agent is None:
                    self.rebel_agent = agent.Agent("", [], verbose=1) # initialize only if tools registered
                self.rebel_agent.set_tools(final_tools)
                use_rebel_agent = True

        try:
            if openai_key is not None:
                openai.api_key = self.openai_key
        except:
            return  {"status":0, "resp":"Issue with OpenAI key"}

        self.optimizer.openai_key = openai.api_key
        # self.agent.set_api_key(openai.api_key,"OPENAI_API_KEY") #
        if os.getenv("OPENAI_API_KEY") is None and use_rebel_agent==True :
            print("warning: you need OPENAI_API_KEY environment variable for ", file=sys.stderr)

        try:
            if not use_rebel_agent:
                completion = self.optimizer.complete(static_context,dynamic_prompt,data_synthesis=data_synthesis,finetune = finetune, **kwargs)
            else:
                completion = self.rebel_agent.run(static_context+dynamic_prompt,[])[0]
        except Exception as e:
            return {"status":0, "resp": str(e)}
        return {"completion":completion, "status": 200}



from llm_vm.utils.typings_llm import *
from llm_vm.agents.FLAT.agent_helper.requests.call_open_ai import call_open_ai

def call_llm(llm_request: LLMCallParams) -> LLMCallReturnType:
    if llm_request['llm'] == LLMCallType.OPENAI_CHAT or llm_request['llm'] == LLMCallType.OPENAI_COMPLETION:
        try:
            return call_open_ai(llm_request)
        except Exception as e:
            print("OpenAI call failed", e, file=sys.stderr)
            return "OpenAI is down!", 0




import sys
from llm_vm.utils.typings_llm import *
import os
import json

def get_newest_decision_model(model: DecisionStep, default_model = OpenAIModel.DAVINCI_TEXT) -> Tuple[str, Union[str, None]]:
    current_dir = os.path.dirname(os.path.abspath(__file__))
    file_name = os.path.join(current_dir, 'models.json')

    model_name = model.value
    default_model_name = default_model.value
    print(file_name, file=sys.stderr)
    file = open(file_name)

    data: List[LLModels] = json.load(file);

    for llm_model in reversed(data):
        if model.value in llm_model and llm_model[model_name]:
            print(f"### {model_name} model will be used (version {llm_model['DATE']})", flush=True, file=sys.stderr)
            return llm_model[model_name]['model_name'], llm_model['OPENAI_KEY']

    print(f"Model {model_name} not found: will use {default_model_name}", flush=True, file=sys.stderr)
    return default_model_name, None



import json
import sys
from sentence_transformers import SentenceTransformer, util
import openai
import backoff 
import os
import time
import pickle
import llm_vm.config as conf
from llm_vm.guided_completion import GenerativeCompletion
import re

class DataSynthesis:
    def __init__(self, variance, examples_to_generate, seed_examples=10):
        self.variance = variance
        self.seed_examples = seed_examples
        self.examples_to_generate = examples_to_generate
        self.call_big = None

    def data_synthesis(self, call_big, prompt, response, openai_key=None, completion=None, seed_batch_size=10, **kwargs):
        """
        This method generates QA pairs using the larger LLM to be used as training data for fine-tuning the smaller LLM.

        Parameters
        ----------
        - optimizer (class): The Optimizer class to use for fine-tuning. Could be either LocalOptimizer or HostedOptimizer.
        - prompt (str | list ): A question to be used as a one-shot QA example for the larger LLM prompt.
        - response (str | list): A verified answer to the provided prompt question to be used in the one-shot QA example.
        - example_delim (str): A unique XML tag used to separate the generated JSON examples. Default value is "<END>".
        - semantic_sim (bool): Option to use semantic similarity to filter duplicate generations. Default value is True.
        - **kwargs: Additional keyword arguments to be passed into the `call_big` method.

        Returns
        ----------
        - List: A list of tuples containing the QA pairs to be used for fine-tuning.
        """
        self.call_big = call_big
        if os.path.isfile(conf.settings.data_gen_file):
            new_file = open(conf.settings.data_gen_file,"rb")
            return list(pickle.load(new_file))
        datapoints_list = []
        final_prompt = '{"prompt": "' + prompt +'"  , "response": "' +response+'" }'
        final_prompt = "Generate {seed_batch_size} json similar to the one below and separate each json using a new line. \n" + final_prompt
        while len(datapoints_list) < self.examples_to_generate:
            remaining_examples = self.examples_to_generate - len(datapoints_list)
            if remaining_examples < seed_batch_size:
                final_prompt = "Generate {remaining_examples} json similar to the one below and separate each json using a new line. \n" + final_prompt                
            datapoints = self.generate_examples(final_prompt, openai_key, completion=completion, call_big_kwargs=kwargs)
            datapoints_list += datapoints
            # Open AI rate limit is 1 request/second 
            time.sleep(1)

        new_file = open(conf.settings.data_gen_file,"wb")
        pickle.dump(datapoints_list, new_file)
        return datapoints_list
    
    @backoff.on_exception(backoff.expo, openai.error.RateLimitError)
    def generate_examples(self, final_prompt, openai_key, example_delim="<END>", model="gpt-4", max_tokens=1000, temperature=1, completion=None, call_big_kwargs={}):
        openai.api_key = openai_key
        cur_prompt = [{'role': "system", 'content': final_prompt}]
        tuple_list = []

        # Generate seed prompts and responses using openAI
        response = openai.ChatCompletion.create(messages=cur_prompt, model=model, max_tokens=max_tokens, temperature=temperature)['choices'][0]['message']['content']
        the_data = response.split("\n")
        # GPT responses might have empty strings
        the_clean_data = []
        if '' in  the_data:
            # loop to remove empty strings
            for i in range(len(the_data)):
                if the_data[i] != '':
                    the_clean_data.append(the_data[i])

        json_list = []
        for data in the_clean_data:
            # Sometimes GPT-4 returns a numbered list
            if data[0] != '{':
                idx = data.index('{')
                json_data = data[idx:]
                json_list.append(json_data)
            else:
                json_list.append(data)
        
        prompt_list = [] 

        # to use open AI prompts + responses seed_examples must be equal to examples_to_generate
        if self.examples_to_generate == self.seed_examples:
            for j in json_list:
                j_dict = json.loads(j)
                clean_res = re.sub('\n', '', j_dict["response"])
                j_tuple = (j_dict["prompt"], clean_res+example_delim)
                tuple_list.append(j_tuple)
            return tuple_list
        else:
            # use seed to generate more examples
            seed_prompts = []
            ex_per_seed = self.examples_to_generate // self.seed_examples
            for j in json_list:
                j_dict = json.loads(j)
                seed_prompts.append(j_dict["prompt"])
            for s in seed_prompts:
                s_examples = []
                while len(s_examples) < ex_per_seed:
                    gen_prompt = f"Generate 1 statement like the one below: \n" + s
                    res = self.call_big(gen_prompt, max_tokens=max_tokens, **call_big_kwargs)
                    s_examples.append(res)
                prompt_list += s_examples

        if completion is None:
            try:
                if len(prompt_list) == 0:
                    for pair in json_list:
                        example = json.loads(pair)
                        prompt = example["prompt"]
                        prompt_list.append(prompt)

                # Use example prompts with big model to generate responses with configured variance                
                for p in prompt_list:
                    completion_response = self.call_big(p, max_tokens=max_tokens, **call_big_kwargs)
                    print(completion_response)
                    the_tuple = (p, completion_response+example_delim)
                    tuple_list.append(the_tuple)
            except Exception as e:
                raise Exception(f"An error has ocurred: {e}")                     
        else:
            try:
                for pair in json_list:
                    example = json.loads(pair)
                    prompt = example["prompt"]
                    prompt_list.append(prompt)
                for p in prompt_list:
                    completion_response = completion.complete(prompt)
                    the_tuple = (p, completion_response+example_delim)
                    tuple_list.append(the_tuple)
            except Exception as e:
                raise Exception(f"An error has ocurred: ${e}")
        # use big model prompts + responses to finetune small model     
        return tuple_list


from llm_vm.utils.typings_llm import *
import json
import os
from llm_vm.agents.FLAT.agent_helper.tool_utils import get_training_tool_subset
from llm_vm.agents.FLAT.agent_helper.tools import GENERIC_TOOLS
from llm_vm.agents.FLAT.models.utils.tool_picker_model.get_training_tools import get_randomised_training_tools
from random import randrange

def __use_tool(tool_id: int, shuffle_value: int = 0, shuffle_modulo: int = 1e9) -> int:
    return int((tool_id + shuffle_value) % shuffle_modulo)

# Get a random subset of the tools, which WILL INCLUDE the given tool_id.
# -> The tools are shuffled to prevent the model from finding patterns
# -> The ids are changed to prevent the model from remembering tools (i.e. Tool=X is always ID=Y)
def __get_random_tool_subset(tool_id: int, shuffle_value: int = 0, shuffle_modulo: int = 1e9) -> str:
    return get_training_tool_subset(
        get_randomised_training_tools(GENERIC_TOOLS, shuffle_value, shuffle_modulo),
        tool_id
    )

def __get_toolpicker_model(tools_input_model: ToolpickerInputModel) -> ToolpickerInputModel:

    current_dir = os.path.dirname(os.path.abspath(__file__))
    file_path = os.path.abspath(os.path.join(current_dir, f'../../raw_data/toolpicker_data.json'))
    json_data: List[QuestionSplitModelJSONData] = json.load(open(file_path, "r"))["data"]

    tools_input_model["data"] = json_data
    complete_tools: List[ToolpickerInputModelData] = []
    for t in tools_input_model["data"]:
        rand_value = randrange(0, len(tools_input_model["data"]))
        answer_key = t["answer"]
        answer_value = ALL_TOOLS_MAP[answer_key].value if answer_key in ALL_TOOLS_MAP else None
        if (answer_value == None):
            raise Exception("Invalid tool name/value: " + answer_key)

        # tool_id_to_use = __use_tool(answer_value, rand_value) if answer_value > 0 else answer_value
        tool_id_to_use = 0
        complete_tools.append({
            "question": t["question"],
            "thought": t["thought"] if "thought" in t else "",
            "mem": [(a[0], a[1]) for a in t["mem"]] if "mem" in t else [],
            "answer": tool_id_to_use,
            "tools": __get_random_tool_subset(tool_id_to_use, rand_value)
        })

    complete_model: ToolpickerInputModel = {
        "openai_model": tools_input_model["openai_model"],
        "data": complete_tools
    }
    return complete_model;

tool_input_data: ToolpickerInputModel = __get_toolpicker_model({
    "openai_model": OpenAIModel.DAVINCI
})



import traceback
import os
import sys
import re
import openai

# Get the current file's directory to grab the python files with common functionality in the utils/ folder
current_dir = os.path.dirname(os.path.abspath(__file__))
grandparent_dir = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
utils_dir = os.path.join(grandparent_dir, 'utils/')
sys.path.append(utils_dir)

from llm_vm.utils.labels import *
import random
from Levenshtein import distance as lev
from llm_vm.utils.typings_llm import *
from bs4 import BeautifulSoup


random_fixed_seed = random.Random(4)

def print_op(*kargs, **kwargs):
    print(*kargs, **kwargs, flush=True, file=sys.stderr)

def verbose_answer(data, answer):
    return f'''<{L_ANSWER_DATA}>{str(data)}</{L_ANSWER_DATA}><{L_ANSWER_SUMMARY}>{answer}</{L_ANSWER_SUMMARY}>'''

def remove_tags_from_html_string(html_string):
    # if not isinstance(html_string, str):
    #     return ""
    # return re.compile(r'<[^>]+>').sub("", html_string), False


    soup = BeautifulSoup(html_string, 'html.parser')
    has_friendly_tags: bool = False

    # Find all tags that are not in the list of tags to preserve
    for tag in soup.find_all(True):
        if tag.name in ['ul', 'ol', 'li', 'a', 'pre', 'code', 'kbd', 'i']:
            has_friendly_tags = True
        elif tag.name in ['style', 'script']:
            tag.extract() # remove tag and content
        else:
            tag.unwrap() # remove tag, but preserve content

    # Get the stripped HTML string
    return str(soup).strip("\n "), has_friendly_tags

def print_big(data, label = ""):
    def do_format(x) -> str:
        formatted_title = "======#====== {:20} ======#======\n"
        if len(x) >= 20:
            return formatted_title.format(x)
        else:
            return formatted_title.format((int((20 - len(x)) / 2) * " ") + x)
    try:
        if len(label):
            print(do_format(str(label).upper()), data, flush=True, file=sys.stderr)
        else:
            print(do_format(str(data)), flush=True, file=sys.stderr)

    except:
        print(label, flush=True, file=sys.stderr)


# Tolerance: If string A can be converted into B in more than {tolerance} steps, they are considered different.
def remove_similars(similars_list, tolerance = 3):
    unique_list = []
    sanitized_list = list(map(lambda item: item.strip(" ").lower(), similars_list))
    similars_len = len(similars_list)

    for i in range(similars_len):
        has_duplicate: bool = False
        for j in range(i+1, similars_len):
            if lev(sanitized_list[i], sanitized_list[j], score_cutoff=tolerance) < tolerance:
                has_duplicate = True
                break
        if has_duplicate == False:
            unique_list.append(similars_list[i])
    return unique_list

def make_interaction_request(human_question, ai_response, data, Q = L_QUESTION, A = L_ANSWER, D = L_ANSWER_DATA, INTERACTION = L_INTERACTION):

    interaction_text = f"<{INTERACTION}>\n<{Q}>{human_question}</{Q}>\n"

    if data and len(data) and isinstance(data, str):
        interaction_text += f'''<{D}>{data}</{D}>'''

    interaction_text += f'''<{A}>{ai_response if ai_response else ''}'''

    stop = f"</{A}>\n</{INTERACTION}>"

    return interaction_text, stop

def make_interaction(human_question, ai_response, data = '',
                     Q = L_QUESTION,
                     A = L_ANSWER,
                     D = L_ANSWER_DATA,
                     INTERACTION = L_INTERACTION):
    text, stop = make_interaction_request(human_question, ai_response, data, Q, A, D, INTERACTION)
    return text + stop

def get_tool_by_id(tools, tool_id):
    for tool in tools:
        if tool['id'] == tool_id:
            return tool
    raise Exception(f"No tool corresponds to id='{tool_id}'")

def tidy_up_subquestions(subquestions_str, main_question):
    sub_questions_raw = subquestions_str.replace("\n", "").split(SUBQ_SPLITTER)

    # First, clear the subquestions of extra symbols that might be confusing
    sub_questions_raw = list(filter(lambda q: len(q) > 1, [q.strip() for q in sub_questions_raw]))
    # then, remove similar subquestions
    unique_subquestions = remove_similars(sub_questions_raw)

    # if we only have one sub question, we don't need to "summarise" at the end => main question == subquestion.
    if len(unique_subquestions) == 1:
        return [main_question], None
    else:
        # to make sure, remove similar subquestions again because we add the main question
        # sub_questions = remove_similars(sub_questions + [question])
        return unique_subquestions, main_question



from llm_vm.agents.FLAT.agent_helper.tool_utils import create_memory_prompt
from llm_vm.agents.FLAT.agent_helper.labels import *
from typings_llm import *
from llm_vm.agents.FLAT.models.utils.answer_from_memory_model.answer_from_memory_model_data import answer_from_memory_data

def __construct_answer_from_memory_jsonl(data: List[QuestionSplitModelData]) -> List[PromptModelEntry]:
    jsonl_entries: List[PromptModelEntry] = []
    for entry in data:
        prompt, stopper = create_memory_prompt([{
            "mem": entry["mem"] if "mem" in entry else None,
            "question": entry["question"],
            "answer": entry["answer"]
        }])

        jsonl_entries.append({
            "prompt": prompt,
            "completion": ("yes" if entry["answer"] else "no") + (stopper if stopper else "")
        })
    return jsonl_entries


def answer_from_memory_jsonl() -> Tuple[PromptModelEntry, str]:
    return (
        __construct_answer_from_memory_jsonl(answer_from_memory_data["data"]),
        answer_from_memory_data["openai_model"].value
    )



from dotenv import load_dotenv
import os
import openai
import sys

from llm_vm.completion.optimize import *

haskell = '''
def call_gpt(state, cur_prompt: str, stop: str, max_tokens = 20, quality = "best", temperature = 0.0):
    if state.verbose > 1:
        print_op("\nGPT input for {" +stop + "} "+ str(len(cur_prompt)) + ".")
    if state.verbose > 2:
        print_op(prepPrintPromptContext(cur_prompt))
    ask_tokens = max_tokens + len(cur_prompt) / 2.7
    if state.verbose > 0:
        print_op("ASK_TOKENS:", ask_tokens)
    if (ask_tokens) > 2049:
        quality = 'best'
    model = { 'best' : ("text-davinci-003", 0.02),
              'okay' : ("text-curie-001", 0.002),
             }[quality]
    def calcCost(p):
        return (len(p) / 2700.0) * model[1]
    cost = calcCost(cur_prompt)
    try:
        ans = openai.Completion.create(
            model=model[0],
            max_tokens=max_tokens,
            stop=stop,
            prompt=cur_prompt,
            temperature=temperature
        )
    except Exception as e:
        print_op("WTF:", e)
        state.price += cost
        return "OpenAI is down!"
    response_text = ans['choices'][0]['text']
    simpleprice = model[1] * ans['usage']['total_tokens'] / 1000
    if state.verbose > 0:
        print_op("SimplePrice: $"+str(simpleprice))
    state.price += simpleprice
    if state.verbose > 2:
        print_op("GPT output:")
        print_op(prepPrintPromptContext(response_text))
        print_op("GPT output fin.\n")
    return response_text
def delete_file(file_name):
    location = os.getcwd()
    path = os.path.join(location, file_name)
    os.remove(path)
    return True
def call_ChatGPT(cur_prompt, stop = None, max_tokens = 20, temperature = 0.2, gpt4 = False):
    ans = openai.ChatCompletion.create(
        model="gpt-3.5-turbo-0301" if not gpt4 else 'gpt-4',
        max_tokens=max_tokens,
        stop=stop,
        messages=cur_prompt,
        temperature=temperature)
    return ans['choices'][0]['message']['content']
    return response_text

def call_gpt(cur_prompt: str, stop: str, max_tokens = 20, quality = "best", temperature = 0.0, model = "text-davinci-003"):
    ans = openai.Completion.create(
        model=model,
        max_tokens=max_tokens,
        stop=stop,
        prompt=cur_prompt,
        temperature=temperature
    )
    return ans['choices'][0]['text']
import gzip
import json
def create_jsonl_file(data_list: list, file_name: str, compress: bool = True) -> None:
    """
    Method saves list of dicts into jsonl file.
    :param data: (list) list of dicts to be stored,
    :param filename: (str) path to the output file. If suffix .jsonl is not given then methods appends
        .jsonl suffix into the file.
    :param compress: (bool) should file be compressed into a gzip archive?
    """
    sjsonl = '.jsonl'
    sgz = '.gz'
    # Check filename
    if not file_name.endswith(sjsonl):
        file_name = file_name + sjsonl
    # Save data
    if compress:
        file_name = file_name + sgz
        with gzip.open(file_name, 'w') as compressed:
            for ddict in data_list:
                jout = json.dumps(ddict) + '\n'
                jout = jout.encode('utf-8')
                compressed.write(jout)
    else:
        with open(file_name, 'w') as out:
            for ddict in data_list:
                jout = json.dumps(ddict) + '\n'
                out.write(jout)
    return file_name, open(file_name, "rb")
'''

def run_test_stub():
    try:
        load_dotenv()
    except:
        pass
    openai_api_key = os.getenv('LLM_VM_OPENAI_API_KEY')
    openai.api_key =openai_api_key
    # anarchy_key = os.getenv('LLM_VM_ANARCHY_KEY')
    print("key:", openai.api_key[0:5], file=sys.stderr)
    optimizer = LocalOptimizer(MIN_TRAIN_EXS=1,openai_key=openai_api_key)
    #optimizer = HostedOptimizer(openai_key = openai.api_key,
    #                            anarchy_key = anarchy_key,
    #                            MIN_TRAIN_EXS=2)
    i = 0
    optimizer.complete("Answer question Q. ","Q: What is the currency in myanmmar", \
                 temperature = 0.0, data_synthesis = True,\
                 min_examples_for_synthesis=0,finetune=True)

if __name__ == "__main__":
    run_test_stub()
    '''
    for h in haskell.splitlines():
        print("At: ", i)
        try:
            print(optimizer.complete("Please convert this line to some haskell:", h + "\nHaskell:", max_tokens = 100, temperature = 0.0))
        except Exception as e:
            print('E:', e)

        time.sleep(2)
        if i > 3 and i < 20:
            time.sleep(120)
        i += 1
    '''



"""
This file contains a list of constants that represent tags for JSON objects.
"""

# REBEL & BACKWARD_CHAINING LABELS
TOOL_ID = "ID"
DESCRIPTION ="DESC"
PARAMS = "PARAMS"

INTERACTION = "QA"
EXAMPLE = "CASE"
CONVERSATION = "MEM"

QUESTION = "Q"
THOUGHT = "THOUGHT"
PROMPT = "P"
RESPONSE = "A"

# FLAT LABELS
L_DESCRIPTION ="DESC"
L_PARAMS = "PARAMS"
L_BOT_INSTRUCTIONS = "SYSTEM"

L_INTERACTION = "INTERACTION"
L_INTERACTIONS = "INTERACTIONS"
L_EXAMPLE = "EXAMPLE"
L_EXAMPLES = "EXAMPLES"
L_CONVERSATION = "CONVERSATION-HISTORY"

L_THOUGHT = "THOUGHT"
L_TOOL_LIST = "TOOLS"
L_TOOL = "TOOL"
L_TOOL_PICKER = "PICK-TOOL"
L_QUESTION = "QUESTION"
L_ANSWER = "ANSWER"
L_ANSWER_DATA = "DATA"
L_ANSWER_SUMMARY = "SUMMARY"

L_SPLIT = "SPLIT"
L_SPLIT_INSTRUCTIONS = "INSTRUCTIONS"

SUBQ_SPLITTER = "|||"

# Human-AI interaction
L_INT_HUMAN = "HUMAN-AI"
# AI-AI interaction
L_INT_AI = "AI-AI"

QUALITY = 0.2






import sys
from dotenv import load_dotenv
import os
import openai
from llm_vm.completion.data_synthesis import DataSynthesis
from llm_vm.completion.optimize import *
from llm_vm.client import Client
import re

client = Client(big_model='gpt', small_model='neo')
if __name__ == "__main__":
    try:
        load_dotenv()
    except:
        pass
    openai.api_key = os.getenv('LLM_VM_OPENAI_API_KEY')
    print("key:", openai.api_key, file=sys.stderr)


    data_synthesizer = DataSynthesis(0.87, 10)

    # for one-shot prompt
    prompt = "What is the currency in myanmmar?"
    response = client.complete(prompt=prompt, context = "", openai_key="", temperature = 0.0)["completion"]
    f_response = re.sub('\n', '', response)
    print(f"Prompt: {prompt} \nResponse: {f_response}", file=sys.stderr)
    lst = data_synthesizer.data_synthesis(client.optimizer, prompt, response, openai_key="", temperature=0.0)
    print("SYNTHESIS COMPLETE:", lst)

    # for k-shot prompt
    # prompt_list = ["What is the currency in Madagascar?", "What is the currency in myanmmar?", "What is the currency in Morocco?"]
    # response_list = []
    # for p in prompt_list:
    #     res = client.complete(prompt=p, context = "", openai_key="", temperature = 0.0)
    #     response_list.append(res["completion"])

    # data_synthesizer.data_synthesis(client.optimizer, prompt_list, response_list, openai_key="", temperature=0.0)
